/*	$Id: score.c,v 1.33 2017/06/07 17:31:46 hacki Exp $ */

/*
 * Copyright (c) 2017 Marcus Glocker <marcus@nazgul.ch>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <fcntl.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "score.h"
#include "spaceinvaders.h"

/*
 * RETURN:
 *	-1 = Could not access score file.
 *	 0 = Found score file.
 *	 1 = Found legacy score file.
 *	 2 = Found invalid score file.
 */
int
score_file_verify(char *filepath)
{
	struct stat sb;

	if (stat(filepath, &sb) == -1)
		return -1;

	if (sb.st_size == SCOREFILE_SIZE)
		return 0;
	if (sb.st_size == SCOREFILE_SIZE_LEGACY)
		return 1;

	return 2;
}

/*
 * RETURN:
 *	-1 = Error.
 *	 0 = Success (sf filled).
 */
int
score_file_read(char *filepath, struct score_file *sf)
{
	int fd, n;

	fd = open(filepath, O_RDONLY);
	if (fd == -1)
		return -1;

	n = read(fd, sf, SCOREFILE_SIZE);
	if (n != SCOREFILE_SIZE) {
		close(fd);
		return -1;
	}

	return 0;
}

/*
 * RETURN:
 *	-1 = Error.
 *	 0 = Success (sf filled).
 */
int
score_file_read_legacy(char *filepath, struct score_file *sf)
{
	int fd, n;
	char hiscore[5];
	char *username;

	fd = open(filepath, O_RDONLY);
	if (fd == -1)
		return -1;

	n = read(fd, hiscore, sizeof(hiscore));
	if (n != SCOREFILE_SIZE_LEGACY) {
		close(fd);
		return -1;
	}

	username = score_get_username();
	strlcpy(sf->username, username, sizeof(sf->username));
	sf->hiscore = ati(hiscore, 0, SCORE_MAX);
	sf->timesec = 0;
	free(username);

	return 0;
}

/*
 * RETURN:
 *	-1 = Error.
 *	 0 = Success (scorefile written).
 */
int
score_file_write(char *filepath, struct score_file *sf)
{
	int fd, n;

	fd = open(filepath, O_CREAT|O_TRUNC|O_WRONLY, 0644);
	if (fd == -1)
		return -1;

	n = write(fd, sf, SCOREFILE_SIZE);
	if (n != SCOREFILE_SIZE) {
		close(fd);
		return -1;
	}

	return 0;
}

/*
 * RETURN:
 * 	Number of valid scores found on the host.
 */
int
score_create_hiscore_list(int entries, struct score_file *sf)
{
	int i, j, nscores;
	char filepath[FILENAME_MAX];
	char *f;
	struct passwd *p;

	if (entries == 1) {
		/* Just create high score for the player. */
		f = score_get_filepath();
		if (score_file_verify(f) != 0) {
			free(f);
			return 0;
		}
		if (score_file_read(f, &sf[0]) != 0) {
			free(f);
			return 0;
		}
		free(f);
		return 1;
	}

	for (p = getpwent(), nscores = 0; p != NULL && nscores < entries;
	     p = getpwent()) {
		snprintf(filepath, sizeof(filepath), "%s/%s",
		    p->pw_dir, SCOREFILE_NAME);
		if (score_file_verify(filepath) != 0)
			continue;
		if (score_file_read(filepath, &sf[nscores]) != 0)
			continue;
		nscores++;
	}
	endpwent();

	if (nscores > 1)
		qsort((void *)sf, nscores, sizeof(*sf), score_cmp);

	/* Remove duplicate entries. */
	for (i = 0, j = 0; j < nscores; j++) {
		if (j > 0) {
			if (strcmp(sf[j-1].username, sf[j].username) == 0)
				continue;
		}
		sf[i++] = sf[j];
	}
	nscores = i;

	return nscores;
}

/*
 * RETURN:
 *	This is the comparsion routine for qsort().
 */
int
score_cmp(const void *a, const void *b)
{
	const struct score_file *c, *d;

	c = a;
	d = b;
	if (c->hiscore < d->hiscore)
		return 1;
	if (c->hiscore == d->hiscore)
		if (c->timesec > d->timesec)
			return 1;

	return 0;
}

void
score_print(void)
{
	int i, n;
	struct score_file sf[HISCORE_USER_MAX];
	char *playtime;

	printf("\nSpace Invaders High Scores\n\n");
	printf("RANK NAME            HI-SCORE   TIME\n");
	printf("-----------------------------------------\n");
	n = score_create_hiscore_list(HISCORE_USER_MAX, sf);
	for (i = 0; i < n && i < HISCORE_LIST_MAX; i++) {
		playtime = score_convert_playtime(sf[i].timesec);
		printf("%02d.  %-14s  %04d       %s\n",
		    i+1, sf[i].username, sf[i].hiscore, playtime);
		free(playtime);
	}
	printf("\n");
	exit(1);
}

/*
 * RETURN:
 *	Pointer to username string.
 */
char *
score_get_username(void)
{
	char *login;
	char username[USERNAME_MAXLEN+1];

	login = getenv("LOGNAME");	
	if (login == NULL) {
		strlcpy(username, "1", sizeof(username));
	} else {
		strlcpy(username, login, sizeof(username));
	}

	return strdup(username);
}

/*
 * RETURN:
 *	Pointer to score filepath string.
 */
char *
score_get_filepath(void)
{
	char *home;
	char filepath[FILENAME_MAX];

	home = getenv("HOME");
	if (home == NULL) {
		return NULL;
	} else {
		snprintf(filepath, sizeof(filepath), "%s/%s",
		    home, SCOREFILE_NAME);
	}

	return strdup(filepath);
}

/*
 * RETURN:
 *	Pointer to score list play time string.
 */
char *
score_convert_playtime(int sec)
{
	int min;
	int hrs;
	char score_time[10];

	min = 0;
	hrs = 0;

	/* Minutes. */
	if (sec >= 60) {
		min = sec / 60;
		sec = sec % 60;
	}

	/* Hours. */
	if (min >= 60) {
		hrs = min / 60;
		min = min % 60;
	}

	snprintf(score_time, sizeof(score_time), "%02dh%02dm%02ds",
	    hrs, min, sec);

	return strdup(score_time);
}

#if !defined(__OpenBSD__) && \
    !defined(__NetBSD__)  && \
    !defined(__FreeBSD__) && \
    !defined(__APPLE__)   && \
    !defined(__sun)
/*
 * Copyright (c) 1998, 2015 Todd C. Miller <Todd.Miller@courtesan.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Copy string src to buffer dst of size dsize.  At most dsize-1
 * chars will be copied.  Always NUL terminates (unless dsize == 0).
 * Returns strlen(src); if retval >= dsize, truncation occurred.
 */
size_t
strlcpy(char *dst, const char *src, size_t dsize)
{
	const char *osrc = src;
	size_t nleft = dsize;

	/* Copy as many bytes as will fit. */
	if (nleft != 0) {
		while (--nleft != 0) {
			if ((*dst++ = *src++) == '\0')
				break;
		}
	}

	/* Not enough room in dst, add NUL and traverse rest of src. */
	if (nleft == 0) {
		if (dsize != 0)
			*dst = '\0';		/* NUL-terminate dst */
		while (*src++)
			;
	}

	return(src - osrc - 1);	/* count does not include NUL */
}
#endif
